﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Transactions;
using HeuristicLab.Services.Hive.Common;
using HeuristicLab.Services.Hive.Common.DataTransfer;

namespace HeuristicLab.Services.Hive {
  public class Hive {
    private static object locker = new object();
    private Dictionary<Guid, DateTime> heartbeats = new Dictionary<Guid, DateTime>();

    /// <summary>
    /// Contains a list job JobIds which have been sent to a slave, but the slave has not yet sent 
    /// a jobProgress of the job with a heartbeat, because he has not finished downloading/deserializing it.
    /// The number value indicates a TimeToLive count that is decremented with each server-heartbeat.
    /// When the number reaches zero, the jobs is assumed to be lost and is set Offline again.
    /// </summary>
    private Dictionary<Guid, int> newAssignedJobs = new Dictionary<Guid, int>();

    private LifecycleManager lifecycleManager;

    private DataAccess.IHiveDao dao { 
      get { return ServiceLocator.Instance.HiveDao; } 
    }
    private DataAccess.IContextFactory<DataAccess.HiveDataContext> contextFactory { 
      get { return ServiceLocator.Instance.ContextFactory; } 
    }
    private IAuthorizationManager auth {
      get { return ServiceLocator.Instance.AuthorizationManager; }
    }

    public Hive() {
      this.lifecycleManager = new LifecycleManager();
      lifecycleManager.ServerHeartbeat += new EventHandler(lifecycleManager_ServerHeartbeat);
    }

    /// <summary>
    /// Check if online slaves send their hearbeats
    /// if not -> set them offline and check if they where calculating a job
    /// </summary>
    void lifecycleManager_ServerHeartbeat(object sender, EventArgs e) {
      lock (locker) {
        Guid[] slaveIds = dao.GetSlaves(x => true).Select(s => s.Id).ToArray();

        foreach (Guid slaveId in slaveIds) {
          if (!heartbeats.ContainsKey(slaveId)) {
            // add a heartbeat NOW and give the slave time to say something for some time (HEARTBEAT_MAX_DIF)
            // otherwise alls the slaves jobs would be aborted, which is not desirable if the server has just been restarted
            heartbeats.Add(slaveId, DateTime.Now);
          } else {
            CheckTimeouts(slaveId);
          }
        }
      }
    }

    private void CheckTimeouts(Guid slaveId) {
      using (TransactionScope scope = new TransactionScope(TransactionScopeOption.Required, new TransactionOptions { IsolationLevel = ApplicationConstants.ISOLATION_LEVEL_SCOPE })) {
        TimeSpan timeSinceLastHeartbeat = DateTime.Now - heartbeats[slaveId];
        if (timeSinceLastHeartbeat.TotalSeconds > ApplicationConstants.HEARTBEAT_TIMEOUT) {
          // slave timed out
          var jobs = dao.GetJobs(x => x.Slave.ResourceId == slaveId);
          foreach (var j in jobs) {
            j.State = JobState.Waiting;
            dao.UpdateJob(j);
          }
          var slave = dao.GetSlave(slaveId);
          if (slave.State != SlaveState.Offline) {
            slave.State = SlaveState.Offline;
            dao.UpdateSlave(slave);
          }
          heartbeats.Remove(slaveId);
        }
        scope.Complete();
      }
    }

  }
}
