﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ServiceModel;
using HeuristicLab.Services.Hive.Common.ServiceContracts;
using HeuristicLab.Services.Hive.Common.DataTransfer;
using HeuristicLab.Hive.Contracts;
using System.IO;
using System.Security.Permissions;
using System.Data.Linq;

namespace HeuristicLab.Services.Hive {


  /// <summary>
  /// Implementation of the Hive service (interface <see cref="IHiveService"/>).
  /// </summary>
  [ServiceBehavior(IncludeExceptionDetailInFaults = true)]
  public class HiveService : IHiveService {
    private DataAccess.IHiveDao dao { 
      get { return ServiceLocator.Instance.HiveDao; } 
    }
    private DataAccess.IContextFactory<DataAccess.HiveDataContext> contextFactory { 
      get { return ServiceLocator.Instance.ContextFactory; } 
    }
    private IAuthorizationManager auth {
      get { return ServiceLocator.Instance.AuthorizationManager; }
    }

    #region Job Methods
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public Guid AddJob(Job jobDto, JobData jobDataDto) {
      using (contextFactory.GetContext()) {
        jobDataDto.JobId = dao.AddJob(jobDto);
        dao.AddJobData(jobDataDto);
        return jobDataDto.JobId;
      }
    }

    public Guid AddChildJob(Guid parentJobId, Job jobDto, JobData jobDataDto) {
      jobDto.ParentJobId = parentJobId;
      return AddJob(jobDto, jobDataDto);
    }

    public Job GetJob(Guid jobId) {
      using (contextFactory.GetContext()) {
        return dao.GetJob(jobId);
      }
    }

    public IEnumerable<Job> GetJobs() {
      using (contextFactory.GetContext()) {
        return dao.GetJobs(x => true);
      }
    }

    public IEnumerable<LightweightJob> GetLightweightJobs(IEnumerable<Guid> jobIds) {
      using (contextFactory.GetContext()) {
        return dao.GetJobs(x => jobIds.Contains(x.JobId)).Select(x => new LightweightJob(x)).ToArray();
      }
    }

    public IEnumerable<LightweightJob> GetLightweightChildJobs(Guid? parentJobId, bool recursive, bool includeParent) {
      using (contextFactory.GetContext()) {
        return GetChildJobs(parentJobId, recursive, includeParent).Select(x => new LightweightJob(x)).ToArray();
      }
    }

    // formerly GetLastSerializedResult
    public JobData GetJobData(Guid jobId) {
      using (contextFactory.GetContext()) {
        return dao.GetJobData(jobId);
      }
    }

    public Stream GetJobDataStreamed(Guid jobId) {
      throw new NotImplementedException();
    }

    public void UpdateJob(Job jobDto, JobData jobDataDto) {
      using (contextFactory.GetContext()) {
        dao.UpdateJob(jobDto);
        dao.UpdateJobData(jobDataDto);
      }
    }

    public void UpdateJobDataStreamed(Stream stream) {
      throw new NotImplementedException();
    }

    public void DeleteChildJobs(Guid jobId) {
      using (contextFactory.GetContext()) {
        var jobs = GetChildJobs(jobId, true, false);
        foreach (var job in jobs) {
          dao.DeleteJob(job.Id);
          dao.DeleteJobData(job.Id);
        };
      }
    }

    public Job AquireJob(Guid slaveId) {
      throw new NotImplementedException();
    }
    #endregion


    #region Job Control Methods
    public void AbortJob(Guid jobId) {
      throw new NotImplementedException();
    }
    public Job PauseJob(JobData serializedJob) {
      throw new NotImplementedException();
    }
    #endregion


    #region HiveExperiment Methods
    public IEnumerable<HiveExperiment> GetHiveExperiments() {
      using (contextFactory.GetContext()) {
        return dao.GetHiveExperiments(x => x.UserId == auth.UserId);
      }
    }

    public void UpdateHiveExperiment(HiveExperiment hiveExperimentDto) {
      using (contextFactory.GetContext()) {
        dao.UpdateHiveExperiment(hiveExperimentDto);
      }
    }

    public void DeleteHiveExperiment(Guid hiveExperimentId) {
      using (contextFactory.GetContext()) {
        HiveExperiment he = dao.GetHiveExperiment(hiveExperimentId);
        if(he.RootJobId.HasValue) {
          var jobs = GetChildJobs(he.RootJobId.Value, true, true);
          foreach(var j in jobs) {
            dao.DeleteJobData(j.Id);
            dao.DeleteJob(j.Id);
          }
        }
        dao.DeleteHiveExperiment(hiveExperimentId);
      }
    }
    #endregion

    #region Project Methods
    public Project GetAllProjects() {
      throw new NotImplementedException();
    }

    public void CreateProject(Project project) {
      throw new NotImplementedException();
    }

    public void ChangeProject(Project project) {
      throw new NotImplementedException();
    }

    public void DeleteProject(Guid projectId) {
      throw new NotImplementedException();
    }

    public IEnumerable<Job> GetJobsByProject(Guid projectId) {
      throw new NotImplementedException();
    }
    #endregion

    #region Login Methods
    public void Login() {
      throw new NotImplementedException();
    }

    public void Login(Slave slave) {
      throw new NotImplementedException();
    }

    public void Logout(Guid clientId) {
      throw new NotImplementedException();
    }
    #endregion

    #region Heartbeat Methods
    public List<MessageContainer> ProcessHeartBeat(HeartBeat hbData) {
      throw new NotImplementedException();
    }
    #endregion

    #region Plugin Methods
    public IEnumerable<HivePluginData> GetPlugins(List<HivePlugin> pluginList) {
      throw new NotImplementedException();
    }

    public System.IO.Stream GetStreamedPlugins(List<HivePlugin> pluginList) {
      throw new NotImplementedException();
    }
    #endregion

    #region Calendar Methods
    public IEnumerable<Appointment> GetCalendar(Guid clientId) {
      throw new NotImplementedException();
    }

    public void SetCalendarStatus(Guid clientId, CalendarState state) {
      throw new NotImplementedException();
    }

    public IEnumerable<Appointment> GetUptimeCalendarForResource(Guid guid) {
      throw new NotImplementedException();
    }

    public void SetUptimeCalendarForResource(Guid guid, IEnumerable<Appointment> appointments, bool isForced) {
      throw new NotImplementedException();
    }
    #endregion

    #region Slave Methods
    public Guid AddSlaveGroup(SlaveGroup slaveGroup) {
      using (contextFactory.GetContext()) {
        return dao.AddSlaveGroup(slaveGroup);
      }
    }

    public IEnumerable<Slave> GetSlaves() {
      using (contextFactory.GetContext()) {
        return dao.GetSlaves(x => true);
      }
    }

    public IEnumerable<SlaveGroup> GetSlaveGroups() {
      using (contextFactory.GetContext()) {
        return dao.GetSlaveGroups(x => true);
      }
    }
    
    public void DeleteSlaveGroup(Guid slaveGroupId) {
      using (contextFactory.GetContext()) {
        dao.DeleteSlaveGroup(slaveGroupId);
      }
    }

    public void AddResourceToGroup(Guid slaveGroupId, Resource resource) {
      using (contextFactory.GetContext()) {
        throw new NotImplementedException();        
      }
    }

    public void RemoveResourceFromGroup(Guid clientGroupId, Guid resourceId) {
      using (contextFactory.GetContext()) {
        throw new NotImplementedException();
      }
    }
    #endregion

    #region Helper Methods
    private IEnumerable<Job> GetChildJobs(Guid? parentJobId, bool recursive, bool includeParent) {
      var jobs = new List<Job>(dao.GetJobs(x => parentJobId == null ? !x.ParentJobId.HasValue : x.ParentJobId.Value == parentJobId));

      if (includeParent) {
        jobs.Add(GetJob(parentJobId.Value));
      }

      if (recursive) {
        var childs = new List<Job>();
        foreach (var job in jobs) {
          childs.AddRange(GetChildJobs(job.Id, recursive, false));
        }
        jobs.AddRange(childs);
      }
      return jobs;
    }
    #endregion


  }
}
